#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
This module implements the handler for state of Denon AVR receivers.

:copyright: (c) 2021 by Oliver Goetz.
:license: MIT, see LICENSE for more details.
"""

import asyncio
import logging
import time
from collections.abc import Hashable
from typing import Optional

import attr

from .appcommand import AppCommandCmdParam, AppCommands
from .const import DENON_ATTR_SETATTR
from .exceptions import AvrCommandError, AvrIncompleteResponseError, AvrProcessingError
from .foundation import DenonAVRFoundation, convert_string_int_bool

_LOGGER = logging.getLogger(__name__)


@attr.s(auto_attribs=True, on_setattr=DENON_ATTR_SETATTR)
class DenonAVRToneControl(DenonAVRFoundation):
    """This class implements tone control functions of Denon AVR receiver."""

    _support_tone_control: Optional[bool] = attr.ib(
        converter=attr.converters.optional(bool), default=None
    )
    _tone_control_status: Optional[bool] = attr.ib(
        converter=attr.converters.optional(convert_string_int_bool), default=None
    )
    _tone_control_adjust: Optional[bool] = attr.ib(
        converter=attr.converters.optional(convert_string_int_bool), default=None
    )
    _bass: Optional[int] = attr.ib(
        converter=attr.converters.optional(int), default=None
    )
    _bass_level: Optional[str] = attr.ib(
        converter=attr.converters.optional(str), default=None
    )
    _treble: Optional[int] = attr.ib(
        converter=attr.converters.optional(int), default=None
    )
    _treble_level: Optional[str] = attr.ib(
        converter=attr.converters.optional(str), default=None
    )
    _setup_lock: asyncio.Lock = attr.ib(default=attr.Factory(asyncio.Lock))

    # Update tags for attributes
    # AppCommand.xml interface
    appcommand_attrs = {AppCommands.GetToneControl: None}

    async def async_setup(self) -> None:
        """Ensure that the instance is initialized."""
        async with self._setup_lock:
            _LOGGER.debug("Starting tone control setup")

            # The first update determines if sound mode is supported
            await self.async_update_tone_control()

            # Add tags for a potential AppCommand.xml update
            if self.support_tone_control:
                for tag in self.appcommand_attrs:
                    self._device.api.add_appcommand_update_tag(tag)

            self._device.telnet_api.register_callback(
                "PS", self._async_sound_detail_callback
            )

            self._is_setup = True
            _LOGGER.debug("Finished tone control setup")

    async def _async_sound_detail_callback(
        self, zone: str, event: str, parameter: str
    ) -> None:
        """Handle a sound detail change event."""
        if self._device.zone != zone:
            return

        if parameter[0:3] == "BAS":
            self._bass = int(parameter[4:]) - 44
            self._bass_level = f"{int(parameter[4:]) - 50}dB"
        elif parameter[0:3] == "TRE":
            self._treble = int(parameter[4:]) - 44
            self._treble_level = f"{int(parameter[4:]) - 50}dB"
        elif parameter == "TONE CTRL OFF":
            self._tone_control_adjust = False
            self._tone_control_status = True
        elif parameter == "TONE CTRL ON":
            self._tone_control_adjust = True
            self._tone_control_status = True

    async def async_update(
        self, global_update: bool = False, cache_id: Optional[Hashable] = None
    ) -> None:
        """Update volume asynchronously."""
        _LOGGER.debug("Starting tone control update")
        # Ensure instance is setup before updating
        if not self._is_setup:
            await self.async_setup()

        # Update state
        await self.async_update_tone_control(
            global_update=global_update, cache_id=cache_id
        )
        _LOGGER.debug("Finished tone control update")

    async def async_update_tone_control(
        self, global_update: bool = False, cache_id: Optional[Hashable] = None
    ):
        """Update tone control status of device."""
        if self._device.use_avr_2016_update is None:
            raise AvrProcessingError(
                "Device is not setup correctly, update method not set"
            )

        if self._is_setup and not self._support_tone_control:
            return

        # Tone control is only available for avr 2016 update
        if not self._device.use_avr_2016_update:
            self._support_tone_control = False
            _LOGGER.info("Tone control not supported")
            return

        try:
            await self.async_update_attrs_appcommand(
                self.appcommand_attrs,
                global_update=global_update,
                cache_id=cache_id,
            )
        except (AvrProcessingError, AvrIncompleteResponseError):
            self._support_tone_control = False
            _LOGGER.info("Tone control not supported")
            return

        if not self._is_setup:
            self._support_tone_control = True
            _LOGGER.info("Tone control supported")

    async def async_set_tone_control_command(
        self, parameter_type: str, value: int
    ) -> None:
        """Post request for tone control commands."""
        cmd = (
            attr.evolve(
                AppCommands.SetToneControl,
                set_command=AppCommandCmdParam(name=parameter_type, text=value),
            ),
        )
        await self._device.api.async_post_appcommand(
            self._device.urls.appcommand, cmd, cache_id=time.time()
        )

    ##############
    # Properties #
    ##############
    @property
    def support_tone_control(self) -> Optional[bool]:
        """Return True if tone control is supported."""
        return self._support_tone_control

    @property
    def tone_control_status(self) -> Optional[bool]:
        """Return value of tone control status."""
        return self._tone_control_status

    @property
    def tone_control_adjust(self) -> Optional[bool]:
        """Return value of tone control adjust."""
        return self._tone_control_adjust

    @property
    def bass(self) -> Optional[int]:
        """Return value of bass."""
        return self._bass

    @property
    def bass_level(self) -> Optional[str]:
        """Return level of bass."""
        return self._bass_level

    @property
    def treble(self) -> Optional[int]:
        """Return value of treble."""
        return self._treble

    @property
    def treble_level(self) -> Optional[str]:
        """Return level of treble."""
        return self._treble_level

    ##########
    # Setter #
    ##########
    async def async_enable_tone_control(self) -> None:
        """Enable tone control to change settings like bass or treble."""
        if self._tone_control_status is None:
            raise AvrCommandError(
                "Cannot enable tone control, Dynamic EQ must be deactivated"
            )

        if self._device.telnet_available:
            telnet_command = self._device.telnet_commands.command_tonecontrol + "ON"
            await self._device.telnet_api.async_send_commands(telnet_command)
        else:
            await self.async_set_tone_control_command("adjust", 1)

    async def async_disable_tone_control(self) -> None:
        """Disable tone control to change settings like bass or treble."""
        if self._tone_control_status is None:
            raise AvrCommandError(
                "Cannot disable tone control, Dynamic EQ must be deactivated"
            )
        if self._device.telnet_available:
            telnet_command = self._device.telnet_commands.command_tonecontrol + "OFF"
            await self._device.telnet_api.async_send_commands(telnet_command)
        else:
            await self.async_set_tone_control_command("adjust", 0)

    async def async_set_bass(self, value: int) -> None:
        """
        Set receiver bass.

        Minimum is 0, maximum at 12

        Note:
        Doesn't work, if Dynamic Equalizer is active.
        """
        if value < 0 or value > 12:
            raise AvrCommandError("Invalid value for bass")
        if self._device.telnet_available:
            if not self.tone_control_adjust:
                await self.async_enable_tone_control()
            telnet_command = self._device.telnet_commands.command_bass + str(value + 44)
            await self._device.telnet_api.async_send_commands(telnet_command)
        else:
            await self.async_enable_tone_control()
            await self.async_set_tone_control_command("bassvalue", value)

    async def async_bass_up(self) -> None:
        """
        Increase level of Bass.

        Note:
        Doesn't work, if Dynamic Equalizer is active.
        """
        if self.bass == 12:
            return
        if self._device.telnet_available:
            if not self.tone_control_adjust:
                await self.async_enable_tone_control()
            telnet_command = self._device.telnet_commands.command_bass + "UP"
            await self._device.telnet_api.async_send_commands(telnet_command)
        else:
            await self.async_enable_tone_control()
            await self.async_set_tone_control_command("bassvalue", self.bass + 1)
            await self.async_update()

    async def async_bass_down(self) -> None:
        """
        Decrease level of Bass.

        Note:
        Doesn't work, if Dynamic Equalizer is active.
        """
        if self.bass == 0:
            return
        if self._device.telnet_available:
            if not self.tone_control_adjust:
                await self.async_enable_tone_control()
            telnet_command = self._device.telnet_commands.command_bass + "DOWN"
            await self._device.telnet_api.async_send_commands(telnet_command)
        else:
            await self.async_enable_tone_control()
            await self.async_set_tone_control_command("bassvalue", self.bass - 1)
            await self.async_update()

    async def async_set_treble(self, value: int) -> None:
        """
        Set receiver treble.

        Minimum is 0, maximum at 12

        Note:
        Doesn't work, if Dynamic Equalizer is active.
        """
        if value < 0 or value > 12:
            raise AvrCommandError("Invalid value for treble")
        if self._device.telnet_available:
            if not self.tone_control_adjust:
                await self.async_enable_tone_control()
            telnet_command = self._device.telnet_commands.command_treble + str(
                value + 44
            )
            await self._device.telnet_api.async_send_commands(telnet_command)
        else:
            await self.async_enable_tone_control()
            await self.async_set_tone_control_command("treblevalue", value)

    async def async_treble_up(self) -> None:
        """
        Increase level of Treble.

        Note:
        Doesn't work, if Dynamic Equalizer is active.
        """
        if self.treble == 12:
            return
        if self._device.telnet_available:
            if not self.tone_control_adjust:
                await self.async_enable_tone_control()
            telnet_command = self._device.telnet_commands.command_treble + "UP"
            await self._device.telnet_api.async_send_commands(telnet_command)
        else:
            await self.async_enable_tone_control()
            await self.async_set_tone_control_command("treblevalue", self.treble + 1)
            await self.async_update()

    async def async_treble_down(self) -> None:
        """
        Decrease level of Treble.

        Note:
        Doesn't work, if Dynamic Equalizer is active.
        """
        if self.treble == 0:
            return
        if self._device.telnet_available:
            if not self.tone_control_adjust:
                await self.async_enable_tone_control()
            telnet_command = self._device.telnet_commands.command_treble + "DOWN"
            await self._device.telnet_api.async_send_commands(telnet_command)
        else:
            await self.async_enable_tone_control()
            await self.async_set_tone_control_command("treblevalue", self.treble - 1)
            await self.async_update()


def tone_control_factory(instance: DenonAVRFoundation) -> DenonAVRToneControl:
    """Create DenonAVRToneControl at receiver instances."""
    # pylint: disable=protected-access
    new = DenonAVRToneControl(device=instance._device)
    return new



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
This module implements the REST API to Denon AVR receivers.

:copyright: (c) 2021 by Oliver Goetz.
:license: MIT, see LICENSE for more details.
"""

import inspect
import logging
import time
from functools import wraps
from typing import Callable, TypeVar

import httpx
from asyncstdlib import lru_cache

from .exceptions import (
    AvrForbiddenError,
    AvrInvalidResponseError,
    AvrNetworkError,
    AvrRequestError,
    AvrTimoutError,
)

_LOGGER = logging.getLogger(__name__)

AnyT = TypeVar("AnyT")


def async_handle_receiver_exceptions(func: Callable[..., AnyT]) -> Callable[..., AnyT]:
    """Handle exceptions raised when calling a Denon AVR endpoint asynchronously."""

    @wraps(func)
    async def wrapper(*args, **kwargs):
        try:
            return await func(*args, **kwargs)
        except httpx.HTTPStatusError as err:
            _LOGGER.debug("HTTP status error on request %s: %s", err.request, err)
            # Separate handling of 403 errors
            if err.response.status_code == 403:
                raise AvrForbiddenError(f"HTTPStatusError: {err}", err.request) from err
            raise AvrRequestError(f"HTTPStatusError: {err}", err.request) from err
        except httpx.TimeoutException as err:
            _LOGGER.debug("HTTP timeout exception on request %s: %s", err.request, err)
            raise AvrTimoutError(f"TimeoutException: {err}", err.request) from err
        except httpx.NetworkError as err:
            _LOGGER.debug("Network error exception on request %s: %s", err.request, err)
            raise AvrNetworkError(f"NetworkError: {err}", err.request) from err
        except httpx.RemoteProtocolError as err:
            _LOGGER.debug(
                "Remote protocol error exception on request %s",
                err.request,
            )
            raise AvrInvalidResponseError(
                f"RemoteProtocolError: {err}", err.request
            ) from err

    return wrapper


def cache_result(func: Callable[..., AnyT]) -> Callable[..., AnyT]:
    """
    Decorate a function to cache its results with an lru_cache of maxsize 32.

    This decorator also sets an "cache_id" keyword argument if it is not set yet.
    """
    if inspect.signature(func).parameters.get("cache_id") is None:
        raise AttributeError(
            f"Function {func} does not have a 'cache_id' keyword parameter"
        )

    lru_decorator = lru_cache(maxsize=32)
    cached_func = lru_decorator(func)

    @wraps(func)
    async def wrapper(*args, **kwargs):
        if kwargs.get("cache_id") is None:
            kwargs["cache_id"] = time.time()

        return await cached_func(*args, **kwargs)

    return wrapper



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
This module implements the handler for volume of Denon AVR receivers.

:copyright: (c) 2021 by Oliver Goetz.
:license: MIT, see LICENSE for more details.
"""

import logging
from collections.abc import Hashable
from typing import Dict, Optional, Union, get_args

import attr

from .appcommand import AppCommands
from .const import (
    CHANNEL_MAP,
    CHANNEL_MAP_LABELS,
    CHANNEL_VOLUME_MAP,
    CHANNEL_VOLUME_MAP_LABELS,
    DENON_ATTR_SETATTR,
    MAIN_ZONE,
    STATE_ON,
    SUBWOOFERS_MAP,
    SUBWOOFERS_MAP_LABELS,
    Channels,
    Subwoofers,
)
from .exceptions import AvrCommandError, AvrProcessingError
from .foundation import DenonAVRFoundation, convert_on_off_bool

_LOGGER = logging.getLogger(__name__)


def convert_muted(value: str) -> bool:
    """Convert muted to bool."""
    return bool(value.lower() == STATE_ON)


def convert_volume(value: Union[float, str]) -> float:
    """Convert volume to float."""
    if value == "--":
        return -80.0
    return float(value)


@attr.s(auto_attribs=True, on_setattr=DENON_ATTR_SETATTR)
class DenonAVRVolume(DenonAVRFoundation):
    """This class implements volume functions of Denon AVR receiver."""

    _volume: Optional[float] = attr.ib(
        converter=attr.converters.optional(convert_volume), default=None
    )
    _muted: Optional[bool] = attr.ib(
        converter=attr.converters.optional(convert_muted), default=None
    )
    _channel_volumes: Optional[Dict[Channels, float]] = attr.ib(default=None)
    _valid_channels = get_args(Channels)
    _subwoofer: Optional[bool] = attr.ib(
        converter=attr.converters.optional(convert_on_off_bool), default=None
    )
    _subwoofer_levels_adjustment: bool = attr.ib(default=True)
    _subwoofer_levels: Optional[Dict[Subwoofers, float]] = attr.ib(default=None)
    _valid_subwoofers = get_args(Subwoofers)
    _lfe: Optional[int] = attr.ib(converter=attr.converters.optional(int), default=None)
    _bass_sync: Optional[int] = attr.ib(
        converter=attr.converters.optional(int), default=None
    )
    # Update tags for attributes
    # AppCommand.xml interface
    appcommand_attrs = {
        AppCommands.GetAllZoneVolume: None,
        AppCommands.GetAllZoneMuteStatus: None,
    }
    # Status.xml interface
    status_xml_attrs = {"_volume": "./MasterVolume/value", "_muted": "./Mute/value"}

    def setup(self) -> None:
        """Ensure that the instance is initialized."""
        # Add tags for a potential AppCommand.xml update
        for tag in self.appcommand_attrs:
            self._device.api.add_appcommand_update_tag(tag)

        self._device.telnet_api.register_callback("MV", self._async_volume_callback)
        self._device.telnet_api.register_callback("MU", self._async_mute_callback)
        self._device.telnet_api.register_callback(
            "CV", self._async_channel_volume_callback
        )
        self._device.telnet_api.register_callback(
            "PS", self._async_subwoofer_state_callback
        )
        self._device.telnet_api.register_callback(
            "PS", self._async_subwoofer_levels_callback
        )
        self._device.telnet_api.register_callback("PS", self._async_lfe_callback)
        self._device.telnet_api.register_callback("PS", self._async_bass_sync_callback)

        self._is_setup = True

    async def _async_volume_callback(
        self, zone: str, event: str, parameter: str
    ) -> None:
        """Handle a volume change event."""
        if self._device.zone != zone:
            return

        if len(parameter) < 3:
            self._volume = -80.0 + float(parameter)
        else:
            whole_number = float(parameter[0:2])
            fraction = 0.1 * float(parameter[2])
            self._volume = -80.0 + whole_number + fraction

    async def _async_mute_callback(self, zone: str, event: str, parameter: str) -> None:
        """Handle a muting change event."""
        if self._device.zone != zone:
            return

        self._muted = parameter

    async def _async_channel_volume_callback(
        self, zone: str, event: str, parameter: str
    ) -> None:
        """Handle a channel volume change event."""
        if event != "CV":
            return

        channel_volume = parameter.split()
        if (
            len(channel_volume) != 2
            or channel_volume[0] not in CHANNEL_MAP_LABELS
            or channel_volume[1] not in CHANNEL_VOLUME_MAP
        ):
            return

        if self._channel_volumes is None:
            self._channel_volumes = {}

        channel = CHANNEL_MAP_LABELS[channel_volume[0]]
        volume = channel_volume[1]
        self._channel_volumes[channel] = CHANNEL_VOLUME_MAP[volume]

    async def _async_subwoofer_state_callback(
        self, zone: str, event: str, parameter: str
    ) -> None:
        """Handle a subwoofer state change event."""
        if parameter[:3] == "SWR":
            self._subwoofer = parameter[4:]

    async def _async_subwoofer_levels_callback(
        self, zone: str, event: str, parameter: str
    ) -> None:
        """Handle a subwoofer levels change event."""
        if parameter[:3] != "SWL":
            return

        subwoofer_volume = parameter.split()
        if (
            len(subwoofer_volume) != 2
            or subwoofer_volume[0] not in SUBWOOFERS_MAP_LABELS
        ):
            return

        if self._subwoofer_levels is None:
            self._subwoofer_levels = {}

        subwoofer = SUBWOOFERS_MAP_LABELS[subwoofer_volume[0]]
        level = subwoofer_volume[1]
        val = convert_on_off_bool(level)
        if val is not None:
            self._subwoofer_levels_adjustment = val
        elif level in CHANNEL_VOLUME_MAP:
            self._subwoofer_levels[subwoofer] = CHANNEL_VOLUME_MAP[level]

    async def _async_lfe_callback(self, zone: str, event: str, parameter: str) -> None:
        """Handle a LFE change event."""
        if parameter[:3] != "LFE":
            return

        self._lfe = int(parameter[4:]) * -1

    async def _async_bass_sync_callback(
        self, zone: str, event: str, parameter: str
    ) -> None:
        """Handle a LFE change event."""
        if parameter[:3] != "BSC":
            return

        self._bass_sync = int(parameter[4:]) * -1

    async def async_update(
        self, global_update: bool = False, cache_id: Optional[Hashable] = None
    ) -> None:
        """Update volume asynchronously."""
        _LOGGER.debug("Starting volume update")
        # Ensure instance is setup before updating
        if not self._is_setup:
            self.setup()

        # Update state
        await self.async_update_volume(global_update=global_update, cache_id=cache_id)
        _LOGGER.debug("Finished volume update")

    async def async_update_volume(
        self, global_update: bool = False, cache_id: Optional[Hashable] = None
    ):
        """Update volume status of device."""
        if self._device.use_avr_2016_update is None:
            raise AvrProcessingError(
                "Device is not setup correctly, update method not set"
            )

        if self._device.use_avr_2016_update:
            await self.async_update_attrs_appcommand(
                self.appcommand_attrs, global_update=global_update, cache_id=cache_id
            )
        else:
            urls = [self._device.urls.status]
            if self._device.zone == MAIN_ZONE:
                urls.append(self._device.urls.mainzone)
            await self.async_update_attrs_status_xml(
                self.status_xml_attrs, urls, cache_id=cache_id
            )

    ##############
    # Properties #
    ##############
    @property
    def muted(self) -> Optional[bool]:
        """
        Boolean if volume is currently muted.

        Return "True" if muted and "False" if not muted.
        """
        return self._muted

    @property
    def volume(self) -> Optional[float]:
        """
        Return volume of Denon AVR as float.

        Volume is send in a format like -50.0.
        Minimum is -80.0, maximum at 18.0
        """
        return self._volume

    @property
    def channel_volumes(self) -> Optional[Dict[Channels, float]]:
        """
        Return the channel levels of the device in dB.

        Only available if using Telnet.
        """
        return self._channel_volumes

    @property
    def subwoofer(self) -> Optional[bool]:
        """
        Return the state of the subwoofer.

        Only available if using Telnet.
        """
        return self._subwoofer

    @property
    def subwoofer_levels(self) -> Optional[Dict[Subwoofers, Union[bool, float]]]:
        """
        Return the subwoofer levels of the device in dB when enabled.

        Only available if using Telnet.
        """
        if self._subwoofer_levels_adjustment:
            return self._subwoofer_levels

        return None

    @property
    def lfe(self) -> Optional[int]:
        """
        Return LFE level in dB.

        Only available if using Telnet.
        """
        return self._lfe

    @property
    def bass_sync(self) -> Optional[int]:
        """
        Return Bass Sync level in dB.

        Only available if using Telnet.
        """
        return self._bass_sync

    ##########
    # Getter #
    ##########
    def channel_volume(self, channel: Channels) -> Optional[float]:
        """
        Return the volume of a channel in dB.

        Only available if using Telnet.
        """
        self._is_valid_channel(channel)
        if self._channel_volumes is None:
            return None
        return self._channel_volumes[channel]

    def subwoofer_level(self, subwoofer: Subwoofers) -> Optional[float]:
        """
        Return the volume of a subwoofer in dB.

        Only available if using Telnet.
        """
        self._is_valid_subwoofer(subwoofer)
        if self._subwoofer_levels is None:
            return None
        return self._subwoofer_levels[subwoofer]

    ##########
    # Setter #
    ##########
    async def async_volume_up(self) -> None:
        """Volume up receiver via HTTP get command."""
        if self._device.telnet_available:
            await self._device.telnet_api.async_send_commands(
                self._device.telnet_commands.command_volume_up, skip_confirmation=True
            )
        else:
            await self._device.api.async_get_command(
                self._device.urls.command_volume_up
            )

    async def async_volume_down(self) -> None:
        """Volume down receiver via HTTP get command."""
        if self._device.telnet_available:
            await self._device.telnet_api.async_send_commands(
                self._device.telnet_commands.command_volume_down, skip_confirmation=True
            )
        else:
            await self._device.api.async_get_command(
                self._device.urls.command_volume_down
            )

    async def async_set_volume(self, volume: float) -> None:
        """
        Set receiver volume via HTTP get command.

        Volume is send in a format like -50.0.
        Minimum is -80.0, maximum at 18.0
        """
        if volume < -80 or volume > 18:
            raise AvrCommandError(f"Invalid volume: {volume}")

        # Round volume because only values which are a multi of 0.5 are working
        volume = round(volume * 2) / 2.0
        if self._device.telnet_available:
            await self._device.telnet_api.async_send_commands(
                self._device.telnet_commands.command_set_volume.format(
                    volume=int(volume + 80)
                )
            )
        else:
            await self._device.api.async_get_command(
                self._device.urls.command_set_volume.format(volume=volume)
            )

    async def async_mute(self, mute: bool) -> None:
        """Mute receiver via HTTP get command."""
        if mute:
            if self._device.telnet_available:
                await self._device.telnet_api.async_send_commands(
                    self._device.telnet_commands.command_mute_on
                )
            else:
                await self._device.api.async_get_command(
                    self._device.urls.command_mute_on
                )
        else:
            if self._device.telnet_available:
                await self._device.telnet_api.async_send_commands(
                    self._device.telnet_commands.command_mute_off
                )
            else:
                await self._device.api.async_get_command(
                    self._device.urls.command_mute_off
                )

    async def async_channel_volume_up(self, channel: Channels) -> None:
        """Increase Channel volume on receiver via HTTP get command."""
        self._is_valid_channel(channel)

        mapped_channel = CHANNEL_MAP[channel]
        if self._device.telnet_available:
            await self._device.telnet_api.async_send_commands(
                self._device.telnet_commands.command_channel_volume.format(
                    channel=mapped_channel, value="UP"
                )
            )
        else:
            await self._device.api.async_get_command(
                self._device.urls.command_channel_volume.format(
                    channel=mapped_channel, value="UP"
                )
            )

    async def async_channel_volume_down(self, channel: Channels) -> None:
        """Decrease Channel volume on receiver via HTTP get command."""
        self._is_valid_channel(channel)

        mapped_channel = CHANNEL_MAP[channel]
        if self._device.telnet_available:
            await self._device.telnet_api.async_send_commands(
                self._device.telnet_commands.command_channel_volume.format(
                    channel=mapped_channel, value="DOWN"
                )
            )
        else:
            await self._device.api.async_get_command(
                self._device.urls.command_channel_volume.format(
                    channel=mapped_channel, value="DOWN"
                )
            )

    async def async_channel_volume(self, channel: Channels, volume: float) -> None:
        """
        Set Channel volume on receiver via HTTP get command.

        :param channel: Channel to set.
        :param volume: Volume to set. Valid values are -12 to 12 with 0.5 steps.
        """
        self._is_valid_channel(channel)
        if volume not in CHANNEL_VOLUME_MAP_LABELS:
            raise AvrCommandError(f"Invalid channel volume: {volume}")

        mapped_channel = CHANNEL_MAP[channel]
        mapped_volume = CHANNEL_VOLUME_MAP_LABELS[volume]
        if self._device.telnet_available:
            await self._device.telnet_api.async_send_commands(
                self._device.telnet_commands.command_channel_volume.format(
                    channel=mapped_channel, value=mapped_volume
                )
            )
        else:
            await self._device.api.async_get_command(
                self._device.urls.command_channel_volume.format(
                    channel=mapped_channel, value=mapped_volume
                )
            )

    async def async_channel_volumes_reset(self) -> None:
        """Reset all channel volumes on receiver via HTTP get command."""
        if self._device.telnet_available:
            await self._device.telnet_api.async_send_commands(
                self._device.telnet_commands.command_channel_volumes_reset
            )
        else:
            await self._device.api.async_get_command(
                self._device.urls.command_channel_volumes_reset
            )

    async def async_subwoofer_on(self) -> None:
        """Turn on Subwoofer on receiver via HTTP get command."""
        if self._device.telnet_available:
            await self._device.telnet_api.async_send_commands(
                self._device.telnet_commands.command_subwoofer_on_off.format(mode="ON")
            )
        else:
            await self._device.api.async_get_command(
                self._device.urls.command_subwoofer_on_off.format(mode="ON")
            )

    async def async_subwoofer_off(self) -> None:
        """Turn off Subwoofer on receiver via HTTP get command."""
        if self._device.telnet_available:
            await self._device.telnet_api.async_send_commands(
                self._device.telnet_commands.command_subwoofer_on_off.format(mode="OFF")
            )
        else:
            await self._device.api.async_get_command(
                self._device.urls.command_subwoofer_on_off.format(mode="OFF")
            )

    async def async_subwoofer_toggle(self) -> None:
        """
        Toggle Subwoofer on receiver via HTTP get command.

        Only available if using Telnet.
        """
        if self._subwoofer:
            await self.async_subwoofer_off()
        else:
            await self.async_subwoofer_on()

    async def async_subwoofer_level_up(self, subwoofer: Subwoofers) -> None:
        """Increase Subwoofer level on receiver via HTTP get command."""
        self._is_valid_subwoofer(subwoofer)
        mapped_subwoofer = SUBWOOFERS_MAP[subwoofer]
        if self._device.telnet_available:
            await self._device.telnet_api.async_send_commands(
                self._device.telnet_commands.command_subwoofer_level.format(
                    number=mapped_subwoofer, mode="UP"
                )
            )
        else:
            await self._device.api.async_get_command(
                self._device.urls.command_subwoofer_level.format(
                    number=mapped_subwoofer, mode="UP"
                )
            )

    async def async_subwoofer_level_down(self, subwoofer: Subwoofers) -> None:
        """Decrease Subwoofer level on receiver via HTTP get command."""
        self._is_valid_subwoofer(subwoofer)
        mapped_subwoofer = SUBWOOFERS_MAP[subwoofer]
        if self._device.telnet_available:
            await self._device.telnet_api.async_send_commands(
                self._device.telnet_commands.command_subwoofer_level.format(
                    number=mapped_subwoofer, mode="DOWN"
                )
            )
        else:
            await self._device.api.async_get_command(
                self._device.urls.command_subwoofer_level.format(
                    number=mapped_subwoofer, mode="DOWN"
                )
            )

    async def async_lfe_up(self) -> None:
        """Increase LFE on receiver via HTTP get command."""
        if self._device.telnet_available:
            await self._device.telnet_api.async_send_commands(
                self._device.telnet_commands.command_lfe.format(mode="UP")
            )
        else:
            await self._device.api.async_get_command(
                self._device.urls.command_lfe.format(mode="UP")
            )

    async def async_lfe_down(self) -> None:
        """Decrease LFE on receiver via HTTP get command."""
        if self._device.telnet_available:
            await self._device.telnet_api.async_send_commands(
                self._device.telnet_commands.command_lfe.format(mode="DOWN")
            )
        else:
            await self._device.api.async_get_command(
                self._device.urls.command_lfe.format(mode="DOWN")
            )

    async def async_lfe(self, lfe: int) -> None:
        """
        Set LFE level on receiver via HTTP get command.

        :param lfe: LFE level to set. Valid values are -10 to 0.
        """
        if lfe < -10 or lfe > 0:
            raise AvrCommandError(f"Invalid LFE: {lfe}")

        lfe_local = str(lfe).replace("-", "").zfill(2)
        if self._device.telnet_available:
            await self._device.telnet_api.async_send_commands(
                self._device.telnet_commands.command_lfe.format(mode=lfe_local)
            )
        else:
            await self._device.api.async_get_command(
                self._device.urls.command_lfe.format(mode=lfe_local)
            )

    async def async_bass_sync_up(self) -> None:
        """Increase Bass Sync on receiver via HTTP get command."""
        if self._device.telnet_available:
            await self._device.telnet_api.async_send_commands(
                self._device.telnet_commands.command_bass_sync.format(mode="UP")
            )
        else:
            await self._device.api.async_get_command(
                self._device.urls.command_bass_sync.format(mode="UP")
            )

    async def async_bass_sync_down(self) -> None:
        """Decrease Bass Sync on receiver via HTTP get command."""
        if self._device.telnet_available:
            await self._device.telnet_api.async_send_commands(
                self._device.telnet_commands.command_bass_sync.format(mode="DOWN")
            )
        else:
            await self._device.api.async_get_command(
                self._device.urls.command_bass_sync.format(mode="DOWN")
            )

    async def async_bass_sync(self, lfe: int) -> None:
        """
        Set Bass Sync level on receiver via HTTP get command.

        :param lfe: Bass Sync level to set. Valid values are -10 to 0.
        """
        if lfe < -10 or lfe > 0:
            raise AvrCommandError(f"Invalid Bass Sync: {lfe}")

        bass_sync_local = str(lfe).replace("-", "").zfill(2)
        if self._device.telnet_available:
            await self._device.telnet_api.async_send_commands(
                self._device.telnet_commands.command_bass_sync.format(
                    mode=bass_sync_local
                )
            )
        else:
            await self._device.api.async_get_command(
                self._device.urls.command_bass_sync.format(mode=bass_sync_local)
            )

    @staticmethod
    def _is_valid_channel(channel: Channels):
        if channel not in DenonAVRVolume._valid_channels:
            raise AvrCommandError("Invalid channel")

    @staticmethod
    def _is_valid_subwoofer(subwoofer: Subwoofers):
        if subwoofer not in DenonAVRVolume._valid_subwoofers:
            raise AvrCommandError("Invalid subwoofer")


def volume_factory(instance: DenonAVRFoundation) -> DenonAVRVolume:
    """Create DenonAVRVolume at receiver instances."""
    # pylint: disable=protected-access
    new = DenonAVRVolume(device=instance._device)
    return new



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
This module implements the Audyssey settings of Denon AVR receivers.

:copyright: (c) 2020 by Oliver Goetz.
:license: MIT, see LICENSE for more details.
"""

import logging
from typing import Optional, get_args

import attr

from .const import (
    DENON_ATTR_SETATTR,
    DIRAC_FILTER_MAP,
    DIRAC_FILTER_MAP_LABELS,
    DiracFilters,
)
from .exceptions import AvrCommandError
from .foundation import DenonAVRFoundation

_LOGGER = logging.getLogger(__name__)


@attr.s(auto_attribs=True, on_setattr=DENON_ATTR_SETATTR)
class DenonAVRDirac(DenonAVRFoundation):
    """Dirac Settings."""

    _dirac_filter: Optional[str] = attr.ib(
        converter=attr.converters.optional(str), default=None
    )
    _dirac_filters = get_args(DiracFilters)

    def setup(self) -> None:
        """Ensure that the instance is initialized."""
        self._device.telnet_api.register_callback(
            "PS", self._async_dirac_filter_callback
        )
        self._is_setup = True

    async def _async_dirac_filter_callback(
        self, zone: str, event: str, parameter: str
    ) -> None:
        """Handle Dirac filter change event."""
        if event == "PS" and parameter[0:5] == "DIRAC":
            self._dirac_filter = DIRAC_FILTER_MAP_LABELS[parameter[6:]]

    ##############
    # Properties #
    ##############

    @property
    def dirac_filter(self) -> Optional[str]:
        """
        Returns the current Dirac filter.

        Only available if using Telnet.

        Possible values are: "Off", "Slot 1", "Slot 2", "Slot 3"
        """
        return self._dirac_filter

    ##########
    # Setter #
    ##########
    async def async_dirac_filter(self, dirac_filter: DiracFilters) -> None:
        """Set Dirac filter."""
        if dirac_filter not in self._dirac_filters:
            raise AvrCommandError("Invalid Dirac filter")

        mapped_filter = DIRAC_FILTER_MAP[dirac_filter]
        if self._device.telnet_available:
            await self._device.telnet_api.async_send_commands(
                self._device.telnet_commands.command_dirac_filter.format(
                    filter=mapped_filter
                )
            )
            return
        await self._device.api.async_get_command(
            self._device.urls.command_dirac_filter.format(filter=mapped_filter)
        )


def dirac_factory(instance: DenonAVRFoundation) -> DenonAVRDirac:
    """Create DenonAVRDirac at receiver instances."""
    # pylint: disable=protected-access
    new = DenonAVRDirac(device=instance._device)
    return new
